/*
* Copyright (c) 2000 - 2023 Samsung Electronics Co., Ltd All Rights Reserved
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/


#include "alarm-manager-db.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sqlite3.h>
#include <glib.h>
#include <tzplatform_config.h>

#include <database.hpp>

extern GSList* g_disabled_alarm_list;
extern __alarm_server_context_t alarm_context;

namespace {

constexpr const char QUERY_CREATE_TABLE_ALARMMGR[] = R"__alarm(
CREATE TABLE IF NOT EXISTS alarmmgr (
  alarm_id INTEGER PRIMARY KEY,
  start INTEGER,
  end INTEGER,
  uid INTEGER,
  global INTEGER,
  is_disabled INTEGER,
  caller_pkgid TEXT,
  callee_pkgid TEXT,
  app_unique_name TEXT,
  app_service_name TEXT,
  app_service_name_mod TEXT,
  bundle TEXT,
  noti_len INTEGER,
  noti TEXT,
  year INTEGER,
  month INTEGER,
  day INTEGER,
  hour INTEGER,
  min INTEGER,
  sec INTEGER,
  msec INTEGER,
  day_of_week INTEGER,
  repeat INTEGER,
  alarm_type INTEGER,
  reserved_info INTEGER,
  dst_service_name TEXT,
  dst_service_name_mod TEXT)
)__alarm";

constexpr const int LATEST_SETTIME_ID = -2;

enum SchemaVersion {
  SCHEMA_VERSION_0 = 0,
  SCHEMA_VERSION_1 = 1,
  SCHEMA_VERSION_LATEST = SCHEMA_VERSION_1
};

struct MigrationFunc {
  const char *change_log;
  int (*func)(void);
};

std::unique_ptr<tizen_base::Database> alarmmgr_db;
bool is_db_corrupted = false;

char* GetCString(int idx, const tizen_base::Database::Result::Record& rec,
    int max_size = -1) {
  std::optional<std::string> str = rec.GetString(idx);
  if (!str)
    return nullptr;
  if (*str == "null")
    return nullptr;

  if (max_size > 0) {
    return strdup(str->substr(max_size).c_str());
  }

  return strdup(str->c_str());
}

int MigrationFuncVersion1() {
  static const char query[] =
    "CREATE TABLE IF NOT EXISTS alarmmgr_temp ("
    " alarm_id INTEGER PRIMARY KEY, "
    " start INTEGER, "
    " end INTEGER, "
    " uid INTEGER, "
    " global INTEGER, "
    " is_disabled INTEGER, "
    " caller_pkgid TEXT, "
    " callee_pkgid TEXT, "
    " app_unique_name TEXT, "
    " app_service_name TEXT, "
    " app_service_name_mod TEXT, "
    " bundle TEXT, "
    " noti_len INTEGER, "
    " noti TEXT, "
    " year INTEGER, "
    " month INTEGER, "
    " day INTEGER, "
    " hour INTEGER, "
    " min INTEGER, "
    " sec INTEGER, "
    " msec INTEGER, "
    " day_of_week INTEGER, "
    " repeat INTEGER, "
    " alarm_type INTEGER, "
    " reserved_info INTEGER, "
    " dst_service_name TEXT, "
    " dst_service_name_mod TEXT "
    ")";

  static const char insert_query[] =
    "INSERT INTO alarmmgr_temp (alarm_id, start, "
    "end, uid, global, is_disabled, caller_pkgid, callee_pkgid, "
    "app_unique_name, app_service_name, app_service_name_mod, "
    "bundle, noti_len, noti, year, month, day, hour, min, sec, msec, "
    "day_of_week, repeat, alarm_type, reserved_info, "
    "dst_service_name, dst_service_name_mod) "
    "SELECT alarm_id, start, end, uid, global, is_disabled, "
    "caller_pkgid, callee_pkgid, app_unique_name, app_service_name, "
    "app_service_name_mod, bundle, noti_len, noti, year, month, "
    "day, hour, min, sec, msec, day_of_week, repeat, alarm_type, "
    "reserved_info, dst_service_name, dst_service_name_mod "
    "FROM alarmmgr";

  LOGD("DB migration init version start");
  alarmmgr_db->OneStepExec({ "DROP TABLE IF EXISTS alarmmgr_temp" });
  alarmmgr_db->OneStepExec({ query });
  alarmmgr_db->OneStepExec({ insert_query });
  alarmmgr_db->OneStepExec({ "DROP TABLE alarmmgr" });
  alarmmgr_db->OneStepExec({ "ALTER TABLE alarmmgr_temp RENAME TO alarmmgr" });

  return ALARMMGR_RESULT_SUCCESS;
}

const MigrationFunc m_func[] = {
  { "SCHEMA_VERSION_0", nullptr },
  { "SCHEMA_VERSION_1", MigrationFuncVersion1 }
};

const char* GetAlarmDbFile() {
  static std::string str;

  if (!str.empty())
    return str.c_str();

  str = tzplatform_mkpath(TZ_SYS_DB, ".alarmmgr.db");
  return str.c_str();
}

int CreateDbVersion() {
  alarmmgr_db->OneStepExec(
      { "CREATE TABLE IF NOT EXISTS schema_version (version INTEGER)" });
  return 0;
}

int UpdateDbVersion(int ver) {
  auto r = alarmmgr_db->Exec({ "DELETE FROM schema_version" });
  if (!r) {
    LOGE("db.Exec() failed: %s", static_cast<const char*>(r));
    return -1;
  }

  auto q = tizen_base::Database::Sql(
      "INSERT INTO schema_version (version) VALUES (?)")
      .Bind(ver);
  r = alarmmgr_db->Exec(q);
  if (!r) {
    LOGE("db.Exec() failed: %s", static_cast<const char*>(r));
    return -1;
  }

  return 0;
}

int GetDbVersion() {
  auto r = alarmmgr_db->Exec({ "SELECT version FROM schema_version" });
  if (!r) {
    LOGE("db.Exec() failed: %s", static_cast<const char*>(r));
    return -1;
  }

  auto rec = r.GetFirstRecord();
  if (!rec)
    return -1;
  return static_cast<int>(rec->Get(0));
}

int DbUpgrade() {
  int ver = GetDbVersion();
  if (ver >= SCHEMA_VERSION_LATEST)
    return 0;

  auto guard = alarmmgr_db->CreateTransactionGuard();
  if (ver == -1) {
    LOGD("Schema version not exists : %d", ver);
    int ret = CreateDbVersion();
    if (ret != 0)
      return ret;

    ver = SCHEMA_VERSION_0;
  }

  for (int i = ver + 1 ; i <= SCHEMA_VERSION_LATEST; i++) {
    if (m_func[i].func() != ALARMMGR_RESULT_SUCCESS)
      return -1;
  }

  int ret = UpdateDbVersion(SCHEMA_VERSION_LATEST);
  guard.Commit();

  return ret;
}

bool DbBusyHandler(int count) {
  if (5 - count > 0) {
    struct timespec time = {
      .tv_sec = 0,
      .tv_nsec = (count + 1) * 100 * 1000 * 1000
    };
    nanosleep(&time, nullptr);
    LOGD("alarmmgr_db: busy handler called. count: %d", count + 1);
    return true;
  } else {
    LOGD("alarmmgr_db: busy handler will return SQLITE_BUSY error");
    return false;
  }
}

bool ReopenDb() {
  unlink(GetAlarmDbFile());
  try {
    alarmmgr_db.reset(
        new tizen_base::Database(
            GetAlarmDbFile(), SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE,
            DbBusyHandler));
    alarmmgr_db->OneStepExec({ QUERY_CREATE_TABLE_ALARMMGR });
  } catch (const tizen_base::DbException& e) {
    LOGE("[ReopenDb] Failed error: %s", e.msg());
    alarmmgr_db.reset();
    return false;
  }

  if (CreateDbVersion() != 0) {
    alarmmgr_db.reset();
    return false;
  }

  if (UpdateDbVersion(SCHEMA_VERSION_LATEST) != 0) {
    alarmmgr_db.reset();
    return false;
  }

  return true;
}

void RecoverDb() {
}

int CheckAlarmMgrTable(const tizen_base::Database& db, bool* is_exist) {
  auto r = db.Exec(
      { "SELECT COUNT(*) FROM sqlite_master WHERE name='alarmmgr'" });
  if (!r) {
    LOGE("db.Exec failed : %s", static_cast<const char*>(r));
    return static_cast<int>(r);
  }

  auto rec = r.GetFirstRecord();
  if (!rec) {
    *is_exist = false;
    return SQLITE_OK;
  }

  int count = static_cast<int>(rec->Get(0));
  LOGD("DB search result[%d]", count);

  if (count > 0)
    *is_exist = true;
  else
    *is_exist = false;

  return SQLITE_OK;
}

int CheckCallback(void* pid, int argc, char** argv, char** notUsed2) {
  if (strcmp(argv[0], "ok") != 0) {
    LOGE("check integrity result : %s" , argv[0]);
    is_db_corrupted = true;
    return -1;
  } else {
    LOGD("check integrity result : %s" , argv[0]);
  }

  return 0;
}

bool SaveAlarms(__alarm_info_t* __alarm_info) {
  base_info_t* base_info = static_cast<base_info_t*>(
      &(__alarm_info->base_info));
  alarm_date_t* start = &base_info->start;
  alarm_mode_t* mode = &base_info->mode;

  auto q = tizen_base::Database::Sql(
      "INSERT INTO alarmmgr (alarm_id, start, end, uid, global, is_disabled, "
      "caller_pkgid, callee_pkgid, app_unique_name, app_service_name, "
      "app_service_name_mod, bundle, noti_len, noti, year, month, day, "
      "hour, min, sec, msec, day_of_week, repeat, alarm_type, reserved_info, "
      "dst_service_name, dst_service_name_mod) "
      "VALUES (?, ?, ?, ?, ?, 0, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, "
      "?, ?, ?, ?, ?, ?)")
      .Bind(__alarm_info->alarm_id)
      .Bind(static_cast<int64_t>(__alarm_info->start))
      .Bind(static_cast<int64_t>(__alarm_info->end))
      .Bind(static_cast<int>(__alarm_info->uid))
      .Bind(__alarm_info->global)
      .Bind(__alarm_info->caller_pkgid)
      .Bind(__alarm_info->callee_pkgid)
      .Bind(__alarm_info->app_unique_name)
      .Bind(__alarm_info->app_service_name)
      .Bind(__alarm_info->app_service_name_mod)
      .Bind(__alarm_info->bundle)
      .Bind(__alarm_info->noti ?
          static_cast<int>(strlen(__alarm_info->noti)) : 0)
      .Bind(__alarm_info->noti)
      .Bind(start->year)
      .Bind(start->month)
      .Bind(start->day)
      .Bind(start->hour)
      .Bind(start->min)
      .Bind(start->sec)
      .Bind(base_info->msec)
      .Bind(mode->u_interval.day_of_week)
      .Bind(mode->repeat)
      .Bind(base_info->alarm_type)
      .Bind(static_cast<int64_t>(base_info->reserved_info))
      .Bind(__alarm_info->dst_service_name)
      .Bind(__alarm_info->dst_service_name_mod);
  auto r = alarmmgr_db->Exec(q);
  if (!r) {
    LOGE("db.Exec failed : %s", static_cast<const char*>(r));
    return false;
  }

  return true;
}

bool UpdateAlarms(__alarm_info_t* __alarm_info) {
  base_info_t* base_info =
      static_cast<base_info_t*>(&(__alarm_info->base_info));
  alarm_date_t* start = &base_info->start;
  alarm_mode_t* mode = &base_info->mode;

  auto q = tizen_base::Database::Sql(
      "UPDATE alarmmgr SET start=?, end=?, uid=?, global=?, is_disabled=0, "
      "caller_pkgid=?, callee_pkgid=?, app_unique_name=?, app_service_name=?, "
      "app_service_name_mod=?, bundle=?, noti_len=?, noti=?, year=?, month=?, "
      "day=?, hour=?, min=?, sec=?, msec=?, day_of_week=?, repeat=?, "
      "alarm_type=?, reserved_info=?, dst_service_name=?, "
      "dst_service_name_mod=? WHERE alarm_id=?")
      .Bind(static_cast<int64_t>(__alarm_info->start))
      .Bind(static_cast<int64_t>(__alarm_info->end))
      .Bind(static_cast<int>(__alarm_info->uid))
      .Bind(__alarm_info->global)
      .Bind(__alarm_info->caller_pkgid)
      .Bind(__alarm_info->callee_pkgid)
      .Bind(__alarm_info->app_unique_name)
      .Bind(__alarm_info->app_service_name)
      .Bind(__alarm_info->app_service_name_mod)
      .Bind(__alarm_info->bundle)
      .Bind(__alarm_info->noti ?
          static_cast<int>(strlen(__alarm_info->noti)) : 0)
      .Bind(__alarm_info->noti)
      .Bind(start->year)
      .Bind(start->month)
      .Bind(start->day)
      .Bind(start->hour)
      .Bind(start->min)
      .Bind(start->sec)
      .Bind(base_info->msec)
      .Bind(mode->u_interval.day_of_week)
      .Bind(mode->repeat)
      .Bind(base_info->alarm_type)
      .Bind(static_cast<int64_t>(base_info->reserved_info))
      .Bind(__alarm_info->dst_service_name)
      .Bind(__alarm_info->dst_service_name_mod)
      .Bind(__alarm_info->alarm_id);

  auto r = alarmmgr_db->Exec(q);
  if (!r) {
    LOGE("db.Exec() failed: %s", static_cast<const char*>(r));
    return false;
  }

  return true;
}

bool DeleteAlarms(alarm_id_t alarm_id) {
  auto q = tizen_base::Database::Sql(
      "DELETE FROM alarmmgr WHERE alarm_id=?")
      .Bind(alarm_id);

  auto r = alarmmgr_db->Exec(q);
  if (!r) {
    LOGE("db.Exec failed : %s", static_cast<const char*>(r));
    return false;
  }

  return true;
}

void LoadAlarmsFromDb() {
  if (!alarmmgr_db)
    return;

  auto r = alarmmgr_db->Exec({ "SELECT * FROM alarmmgr" });
  if (!r) {
    LOGE("db.Exec() failed: %s", static_cast<const char*>(r));
    return;
  }

  for (const auto& rec : r) {
    int col_idx = 0;
    __alarm_info_t* __alarm_info = reinterpret_cast<__alarm_info_t*>(
        calloc(1, sizeof(__alarm_info_t)));

    if (G_UNLIKELY(__alarm_info == nullptr)) {
      LOGE("Memory allocation failed.");
      break;
    }

    base_info_t* base_info = (base_info_t *) &(__alarm_info->base_info);
    alarm_date_t* start = &base_info->start;
    alarm_mode_t* mode = &base_info->mode;
    __alarm_info->alarm_id = static_cast<int>(rec.Get(col_idx++));
    int64_t start_64 = static_cast<int64_t>(rec.Get(col_idx++));
    __alarm_info->start = static_cast<time_t>(start_64);
    int64_t end_64 = static_cast<int64_t>(rec.Get(col_idx++));
    __alarm_info->end = static_cast<time_t>(end_64);
    __alarm_info->uid = static_cast<int>(rec.Get(col_idx++));
    __alarm_info->global = static_cast<int>(rec.Get(col_idx++));
    int is_disabled = static_cast<int>(rec.Get(col_idx++));
    __alarm_info->caller_pkgid = GetCString(col_idx++, rec);
    __alarm_info->callee_pkgid = GetCString(col_idx++, rec);
    __alarm_info->app_unique_name = GetCString(col_idx++, rec);
    __alarm_info->app_service_name = GetCString(col_idx++, rec);
    __alarm_info->app_service_name_mod = GetCString(col_idx++, rec);
    __alarm_info->bundle = GetCString(col_idx++, rec);
    int noti_len = static_cast<int>(rec.Get(col_idx++));
    noti_len = noti_len ? noti_len : strlen("null");
    __alarm_info->noti = GetCString(col_idx++, rec, noti_len);
    start->year = static_cast<int>(rec.Get(col_idx++));
    start->month = static_cast<int>(rec.Get(col_idx++));
    start->day = static_cast<int>(rec.Get(col_idx++));
    start->hour = static_cast<int>(rec.Get(col_idx++));
    start->min = static_cast<int>(rec.Get(col_idx++));
    start->sec = static_cast<int>(rec.Get(col_idx++));
    base_info->msec = static_cast<int>(rec.Get(col_idx++));
    mode->u_interval.day_of_week = static_cast<int>(rec.Get(col_idx++));
    mode->repeat = static_cast<alarm_repeat_mode_t>(static_cast<int>(rec.Get(col_idx++)));
    base_info->alarm_type = static_cast<int>(rec.Get(col_idx++));
    int64_t reserved_info_64 = static_cast<int64_t>(rec.Get(col_idx++));
    base_info->reserved_info = static_cast<time_t>(reserved_info_64);
    __alarm_info->dst_service_name = GetCString(col_idx++, rec);
    __alarm_info->dst_service_name_mod = GetCString(col_idx++, rec);

    if (is_disabled) {
      _alarm_set_next_duetime(__alarm_info);
      g_disabled_alarm_list = g_slist_append(g_disabled_alarm_list, __alarm_info);
      LOGW("Save alarm_id[%d] caller[%s] callee[%s]", __alarm_info->alarm_id,
          __alarm_info->caller_pkgid, __alarm_info->callee_pkgid);
    } else {
      _alarm_set_next_duetime(__alarm_info);
      alarm_context.alarms = g_slist_append(alarm_context.alarms, __alarm_info);
    }
  }

  _alarm_schedule();
}

int GetDbPathForAllInfo(uid_t uid, char** db_path) {
  char db_path_tmp[50] = {0,};
  constexpr const char query_for_creating_table[] =
      "CREATE TABLE alarmmgr_tool (alarm_id integer primary key, "
      "duetime_epoch integer, duetime text, start_epoch integer, "
      "end_epoch integer, global integer, caller_pkgid text, "
      "callee_pkgid text, app_unique_name text, app_service_name text, "
      "dst_service_name text, day_of_week integer, repeat integer, "
      "alarm_type integer)";
  constexpr const char query_for_deleting_table[] =
      "DROP TABLE alarmmgr_tool";
  time_t current_time = 0;
  struct tm current_tm;
  int index = 0;

  time(&current_time);
  tzset();
  localtime_r(&current_time, &current_tm);
  snprintf(db_path_tmp, sizeof(db_path_tmp),
      "/tmp/alarmmgr_%d%d%d_%02d%02d%02d.db",
      current_tm.tm_year + 1900, current_tm.tm_mon + 1,
      current_tm.tm_mday, current_tm.tm_hour, current_tm.tm_min,
      current_tm.tm_sec);

  *db_path = strdup(db_path_tmp);
  if (*db_path == nullptr) {
    LOGE("Out of memory");
    return ERR_ALARM_SYSTEM_FAIL;
  }

  tizen_base::Database alarmmgr_tool_db(db_path_tmp,
      SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE,
      DbBusyHandler);
  auto r = alarmmgr_tool_db.Exec({ query_for_deleting_table });
  if (!r)
    LOGE("Deleting the table failed : %s", static_cast<const char*>(r));

  alarmmgr_tool_db.OneStepExec({ query_for_creating_table });
  for (GSList* gs_iter = alarm_context.alarms; gs_iter != nullptr;
      gs_iter = g_slist_next(gs_iter)) {
    __alarm_info_t* entry = reinterpret_cast<__alarm_info_t*>(
        gs_iter->data);
    if (uid >= REGULAR_UID_MIN && entry->uid != uid)
      continue;
    ++index;
    SECURE_LOGD("#%d alarm id[%d] app_name[%s] duetime[%ld]",
        index, entry->alarm_id, entry->app_unique_name, entry->start);

    base_info_t *base_info = (base_info_t *) &(entry->base_info);
    alarm_mode_t *mode = &base_info->mode;

    auto q = tizen_base::Database::Sql(
        "INSERT INTO alarmmgr_tool( alarm_id, duetime_epoch, duetime, "
        "start_epoch, end_epoch, global, caller_pkgid, callee_pkgid, "
        "app_unique_name, app_service_name, dst_service_name, day_of_week, "
        "repeat, alarm_type) "
        "VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)")
        .Bind(entry->alarm_id)
        .Bind(static_cast<int>(entry->due_time))
        .Bind(ctime(&(entry->due_time)))
        .Bind(static_cast<int>(entry->start))
        .Bind(static_cast<int>(entry->end))
        .Bind((bool)entry->global)
        .Bind(entry->caller_pkgid)
        .Bind(entry->callee_pkgid)
        .Bind(entry->app_unique_name)
        .Bind(entry->app_service_name)
        .Bind(entry->dst_service_name)
        .Bind(mode->u_interval.day_of_week)
        .Bind(mode->repeat)
        .Bind(entry->base_info.alarm_type);
    auto r = alarmmgr_tool_db.Exec(q);
    if (!r)
      LOGE("db.Exec failed : %s", static_cast<const char*>(r));
  }

  return ALARMMGR_RESULT_SUCCESS;
}

void UpdateRelativeAlarms(gpointer user_data) {
  auto guard = alarmmgr_db->CreateTransactionGuard();

  for (GSList* iter = alarm_context.alarms; iter != nullptr;
      iter = g_slist_next(iter)) {
    __alarm_info_t* entry = reinterpret_cast<__alarm_info_t*>(iter->data);
    base_info_t* base_info = &(entry->base_info);
    if (base_info->alarm_type & ALARM_TYPE_RELATIVE)
      _update_alarms(entry);
  }

  guard.Commit();
}

bool AlarmSetGlobalToDb(__alarm_info_t* alarm_info, bool global) {
  auto q = tizen_base::Database::Sql(
      "UPDATE alarmmgr SET global=? WHERE alarm_id=?")
      .Bind(alarm_info->global)
      .Bind(alarm_info->alarm_id);

  auto r = alarmmgr_db->Exec(q);
  if (!r) {
    LOGE("db.Exec failed : %s", static_cast<const char*>(r));
    return false;
  }

  return true;
}

void UpdateDbForDisabledAlarm(alarm_id_t alarm_id, bool disabled) {
  LOGW("Update (%d) is_disabled to (%d)", alarm_id, disabled);
  auto q = tizen_base::Database::Sql(
      "UPDATE alarmmgr SET is_disabled=? WHERE alarm_id=?")
      .Bind(disabled)
      .Bind(alarm_id);

  auto r = alarmmgr_db->Exec(q);
  if (!r)
    LOGE("db.Exec failed : %s", static_cast<const char*>(r));
}

bool InitializeDb() {
  try {
    alarmmgr_db.reset(new tizen_base::Database(GetAlarmDbFile(),
        SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, DbBusyHandler));
  } catch (const tizen_base::DbException& e) {
    LOGE("Failed to open [%s]. error: %s", GetAlarmDbFile(), e.msg());
    if (e.code() == SQLITE_CORRUPT)
      RecoverDb();
    else
      return false;
  }

  bool is_exist = false;
  int ret = CheckAlarmMgrTable(*alarmmgr_db, &is_exist);
  if (ret != SQLITE_OK) {
    if (ret == SQLITE_CORRUPT) {
      RecoverDb();
    } else {
      alarmmgr_db.reset();
      return false;
    }
  }

  auto r = alarmmgr_db->Exec({QUERY_CREATE_TABLE_ALARMMGR});

  if (!r) {
    LOGE("Don't execute query = %s, error message = %s",
        QUERY_CREATE_TABLE_ALARMMGR, static_cast<const char*>(r));
    if (static_cast<int>(r) == SQLITE_CORRUPT ||
        static_cast<int>(r) == SQLITE_NOTADB) {
      RecoverDb();
      alarmmgr_db.reset();
    }

    return false;
  }

  if (is_exist == false) {
    if (CreateDbVersion() != 0) {
      alarmmgr_db.reset();
      return false;
    }

    if (UpdateDbVersion(SCHEMA_VERSION_LATEST) != 0) {
      alarmmgr_db.reset();
      return false;
    }
  } else {
    ret = DbUpgrade();
    if (ret) {
      LOGE("Failed db's migration");
      alarmmgr_db.reset();
      return false;
    }
  }

  ret = sqlite3_exec(alarmmgr_db->GetRaw(), "PRAGMA integrity_check",
      CheckCallback, nullptr, 0);
  if (ret != SQLITE_OK || is_db_corrupted) {
    LOGE("Loss alarm db's integrity");
    return ReopenDb();
  }

  return true;
}

bool SetLatestSettime(time_t settime) {
  auto q = tizen_base::Database::Sql(
      "INSERT OR REPLACE INTO alarmmgr( alarm_id, start, "
      "end, uid, global, is_disabled, caller_pkgid, callee_pkgid, "
      "app_unique_name, app_service_name, app_service_name_mod, bundle, "
      "noti_len, noti, year, month, day, hour, min, sec, msec, day_of_week, "
      "repeat, alarm_type, reserved_info, dst_service_name, "
      "dst_service_name_mod) "
      "VALUES (?, ?, 0, 0, 0, 1, 'n', 'n', 'n', 'n', 'n', 'n', 0, 'n', "
      "0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 'n', 'n')")
      .Bind(LATEST_SETTIME_ID)
      .Bind(static_cast<int64_t>(settime));

  if (!alarmmgr_db)
    return false;

  auto r = alarmmgr_db->Exec(q);
  if (!r) {
    LOGE("db.Exec failed : %s", static_cast<const char*>(r));
    return false;
  }

  LOGD("latest_settime : %ld", settime);
  return true;
}

bool GetLatestSettime(time_t* settime) {
  auto q = tizen_base::Database::Sql(
      "SELECT start FROM alarmmgr WHERE alarm_id=?")
      .Bind(LATEST_SETTIME_ID);

  if (!alarmmgr_db)
    return false;

  auto r = alarmmgr_db->Exec(q);
  if (!r) {
    LOGE("db.Exec failed : %s", static_cast<const char*>(r));
    return false;
  }

  auto rec = r.GetFirstRecord();
  if (!rec) {
    LOGE("latest_settime is not set");
    return false;
  }

  *settime = static_cast<int64_t>(rec->Get(0));
  LOGD("latest_settime : %ld", *settime);
  return true;
}

}  // namespace

bool _save_alarms(__alarm_info_t* __alarm_info) {
  try {
    return SaveAlarms(__alarm_info);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return false;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return false;
  }
}

bool _update_alarms(__alarm_info_t* __alarm_info) {
  try {
    return UpdateAlarms(__alarm_info);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return false;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return false;
  }
}

bool _delete_alarms(alarm_id_t alarm_id) {
  try {
    return DeleteAlarms(alarm_id);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return false;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return false;
  }
}

void _load_alarms_from_db() {
  try {
    LoadAlarmsFromDb();
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
  }
}

int _get_db_path_for_all_info(uid_t uid, char** db_path) {
  try {
    return GetDbPathForAllInfo(uid, db_path);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return ERR_ALARM_SYSTEM_FAIL;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return ERR_ALARM_SYSTEM_FAIL;
  }
}

gboolean _update_relative_alarms(gpointer user_data) {
  try {
    UpdateRelativeAlarms(user_data);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
  }

  return G_SOURCE_REMOVE;
}

bool _alarm_set_global_to_db(__alarm_info_t* alarm_info, bool global) {
  try {
    return AlarmSetGlobalToDb(alarm_info, global);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return false;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return false;
  }
}

void _update_db_for_disabled_alarm(alarm_id_t alarm_id, bool disabled) {
  try {
    UpdateDbForDisabledAlarm(alarm_id, disabled);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
  }
}

bool _initialize_db() {
  try {
    return InitializeDb();
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return false;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return false;
  }
}

bool _set_latest_settime(time_t settime) {
  try {
    return SetLatestSettime(settime);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return false;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return false;
  }
}

bool _get_latest_settime(time_t* settime) {
  try {
    return GetLatestSettime(settime);
  } catch (const tizen_base::DbException& e) {
    LOGE("Exception: %s", e.msg());
    return false;
  } catch (const std::exception& e) {
    LOGE("Exception: %s", e.what());
    return false;
  }
}
