/*
 * Copyright (c) 2020 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <dlog.h>
#include <sys/ioctl.h>
#include <linux/rtc.h>

#include "plugin.h"

static const char default_rtc[] = "/dev/rtc";
static int gfd = -1;
static int is_psm = -1;

EXPORT int CLEAR_WAKEUP_RTC(void)
{
	struct rtc_wkalrm rtc_wkalarm;

	rtc_wkalarm.enabled = 0;
	rtc_wkalarm.time.tm_year = 1900;
	rtc_wkalarm.time.tm_mon = 0;
	rtc_wkalarm.time.tm_mday = 1;
	rtc_wkalarm.time.tm_hour = 0;
	rtc_wkalarm.time.tm_min = 0;
	rtc_wkalarm.time.tm_sec = 0;

	return ioctl(gfd, RTC_WKALM_SET, &rtc_wkalarm);
}

EXPORT int SET_WAKEUP_RTC(time_t due_time)
{
	struct rtc_wkalrm rtc_wkalarm;
	struct tm due_tm;

	gmtime_r(&due_time, &due_tm);
	LOGD("Setted RTC Alarm date/time is %d-%d-%d, %02d:%02d:%02d (UTC).",
		due_tm.tm_mday, due_tm.tm_mon + 1, due_tm.tm_year + 1900,
		due_tm.tm_hour, due_tm.tm_min, due_tm.tm_sec);

	rtc_wkalarm.enabled = 1;
	rtc_wkalarm.time.tm_year = due_tm.tm_year;
	rtc_wkalarm.time.tm_mon = due_tm.tm_mon;
	rtc_wkalarm.time.tm_mday = due_tm.tm_mday;
	rtc_wkalarm.time.tm_hour = due_tm.tm_hour;
	rtc_wkalarm.time.tm_min = due_tm.tm_min;
	rtc_wkalarm.time.tm_sec = due_tm.tm_sec;

	return ioctl(gfd, RTC_WKALM_SET, &rtc_wkalarm);
}

EXPORT int SET_WAKEUP_RTC_WITH_DATETIME(struct tm *tm)
{
	struct rtc_wkalrm rtc_wkalarm;

	rtc_wkalarm.enabled = 1;
	rtc_wkalarm.time.tm_year = tm->tm_year;
	rtc_wkalarm.time.tm_mon = tm->tm_mon;
	rtc_wkalarm.time.tm_mday = tm->tm_mday;
	rtc_wkalarm.time.tm_hour = tm->tm_hour;
	rtc_wkalarm.time.tm_min = tm->tm_min;
	rtc_wkalarm.time.tm_sec = tm->tm_sec;

	return ioctl(gfd, RTC_WKALM_SET, &rtc_wkalarm);
}

EXPORT int SET_SYSTIME_RTC(time_t new_time)
{
	struct tm tm, *gmtime_res;
	struct rtc_time _rtc_time;

	gmtime_res = gmtime_r(&(new_time), &tm);
	if (!gmtime_res) {
		LOGE("gmtime_r is failed. [%d]", errno);
		return -1;
	}

	_rtc_time.tm_sec = tm.tm_sec;
	_rtc_time.tm_min = tm.tm_min;
	_rtc_time.tm_hour = tm.tm_hour;
	_rtc_time.tm_mday = tm.tm_mday;
	_rtc_time.tm_mon = tm.tm_mon;
	_rtc_time.tm_year = tm.tm_year;
	_rtc_time.tm_wday = tm.tm_wday;
	_rtc_time.tm_yday = tm.tm_yday;
	_rtc_time.tm_isdst = tm.tm_isdst;

	return ioctl(gfd, RTC_SET_TIME, &_rtc_time);
}

EXPORT void DO_AFTER_TIMEZONE_SET(void)
{
	return;
}

EXPORT void DO_AFTER_ALARM_CREATE(int pid, const char *unique_name)
{
	return;
}

EXPORT void DO_AFTER_ALARM_EXPIRE(int pid, const char *unique_name)
{
	return;
}

EXPORT int GET_POWER_SAVING_MODE(void)
{
	return is_psm;
}

EXPORT int RTC_INIT(void)
{
	const char *rtc = default_rtc;

	if (gfd < 0) {
		gfd = open(rtc, O_RDWR);
		if (gfd < 0) {
			LOGE("RTC open failed.");
			return -1;
		}
	}

	return 0;
}

EXPORT int PS_MODE_INIT(void)
{
	return 0;
}
