/*
 * Copyright (c) 2021 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <appsvc.h>
#include <bundle_cpp.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <stdio.h>
#include <unistd.h>

#include "alarm.h"
#include "alarm-internal.h"
#include "aul_mock.h"
#include "dbus_mock.h"
#include "pkgmgr_info_mock.h"
#include "test_fixture.h"
#include "tzplatform_config_mock.h"
#include "unittest.h"

using ::testing::_;
using ::testing::DoAll;
using ::testing::Return;
using ::testing::SetArgPointee;
using ::testing::Invoke;

namespace {

GDBusConnection* __g_bus_get_sync_fake(GBusType type, GCancellable* cancel,
    GError** error) {
  GDBusConnection* con = (GDBusConnection*)g_object_new(G_TYPE_OBJECT,
      nullptr);
  return con;
}

void __g_dbus_node_info_unref_fake(GDBusNodeInfo* info) {
   if (info && info->interfaces) {
     free(info->interfaces);
     free(info);
   }
  info = nullptr;
  return;
}

GDBusNodeInfo* __g_dbus_node_info_new_for_xml_fake(const gchar* xml,
    GError** error) {
  GDBusNodeInfo* info = (GDBusNodeInfo*)malloc(sizeof(GDBusNodeInfo));
  if (!info)
    return nullptr;

  GDBusInterfaceInfo** interface =
    (GDBusInterfaceInfo**)malloc(sizeof(GDBusInterfaceInfo*));
  if (!interface) {
    free(info);
    return nullptr;
  }

  info->ref_count = 10;
  info->path = nullptr;
  info->interfaces = interface;
  info->nodes = nullptr;
  info->annotations = nullptr;

  return info;
}

class Mocks : virtual public ::testing::NiceMock<AulMock>,
    virtual public ::testing::NiceMock<DbusMock>,
    virtual public ::testing::NiceMock<PkgmgrInfoMock>,
    virtual public ::testing::NiceMock<TzplatformConfigMock> {};
}  // namespace

class AlarmServerTest : public TestFixture {
 public:
  AlarmServerTest() : TestFixture(std::make_unique<::Mocks>()) {}

  virtual void SetUp() {}
  virtual void TearDown() {}
};

TEST_F(AlarmServerTest, alarm_manager_alarm_create_p) {
  EXPECT_CALL(GetMock<TzplatformConfigMock>(), tzplatform_mkpath(_, _))
      .WillRepeatedly(Invoke([](enum tzplatform_variable id,
          const char* path) {
        return ".alarm_test.db";
      }));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_signal_subscribe(_, _, _,
       _, _, _, _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_new_for_xml(_, _))
      .WillRepeatedly(Invoke(__g_dbus_node_info_new_for_xml_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_unref(_))
      .WillRepeatedly(Invoke(__g_dbus_node_info_unref_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_register_object(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_own_name_on_connection(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));

  _alarm_initialize();
  int alarm_id = 0;
  gint64 reserved = 1;
  GVariant* param = g_variant_new("(ssiiiiiiiiiiiiixss)",
      "app_service_name", "app_service_name_mod",
      2019, 11, 1, 12, 59, 59, 400, 2020, 12, 29, 3, 1, 1, reserved,
      "reserved_service_name", "reserved_service_name_mod");
  int ret = alarm_manager_alarm_create(param, 5001, getpid(), "sender",
      &alarm_id);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_create_appsvc_p) {
  EXPECT_CALL(GetMock<PkgmgrInfoMock>(),
      pkgmgrinfo_pkginfo_get_api_version(_, _))
      .WillRepeatedly(Invoke([](pkgmgrinfo_pkginfo_h handle,
          char** version) -> int {
        static char ver[] = "2.0";
        *version = ver;
        return 0;
      }));
  EXPECT_CALL(GetMock<PkgmgrInfoMock>(), pkgmgrinfo_appinfo_get_pkgid(_, _))
      .WillRepeatedly(Invoke([](pkgmgrinfo_appinfo_h handle,
          char **pkgid) -> int {
        static char sPkgid[] = "org.tizen.alarm_unitest";
        *pkgid = sPkgid;
        return 0;
      }));

  gint64 reserved = 1;
  tizen_base::Bundle b = { {"__APP_SVC_PKG_NAME__", "org.tizen.alarm" } };
  auto raw = b.ToRaw();
  GVariant* param = g_variant_new("(iiiiiiiiiixiixs)",
      2019, 11, 1, 12, 59, 59,
      2020, 12, 29,
      1, reserved, 1, 0, reserved,
      reinterpret_cast<char*>(raw.first.get()));
  int alarm_id = 0;
  int ret = alarm_manager_alarm_create_appsvc(param, 5001, getpid(), "sender",
      &alarm_id);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_create_periodic_p) {
  GVariant* param = g_variant_new("(ssiii)",
      "app_service_name", "app_service_name_mod", 1, 0, 1);

  int alarm_id = 0;
  int ret = alarm_manager_alarm_create_periodic(param, 5001, getpid(), "sender",
      &alarm_id);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_create_noti_p) {
  int alarm_id;
  gint64 reserved = 1;
  GVariant* param = g_variant_new("(iiiiiiiiiixiixs)",
      2019, 11, 1, 12, 59, 59,
      2020, 12, 29, 3,
      reserved, 1, 0, reserved,
      "bundle");
  int ret = alarm_manager_alarm_create_noti(param, 5001, getpid(), "sender",
      &alarm_id);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_delete_p) {
  GVariant* param = g_variant_new("(i)", 1);
  int ret = alarm_manager_alarm_delete(param, 5001, getpid(), "sender");
  EXPECT_EQ(ERR_ALARM_INVALID_ID, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_delete_all_p) {
  GVariant* param = nullptr;
  int ret = alarm_manager_alarm_delete_all(param, 5001, getpid(), "sender");
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_update_n) {
  int alarm_id = 1;
  gint64 reserved = 1;
  GVariant* param = g_variant_new("(iiiiiiiiiixiixi)",
      alarm_id,
      2019, 11, 1, 12, 59, 59,
      2020, 12, 29, 0,
      reserved, 0, 0, reserved, 0);
  int ret = alarm_manager_alarm_update(param, 5001, getpid(), "sender");
  EXPECT_EQ(ERR_ALARM_INVALID_ID, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_number_of_ids_p) {
  int num_of_ids = 0;
  int ret = alarm_manager_alarm_get_number_of_ids(5001, getpid(), "sender",
      &num_of_ids);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_list_of_ids_p) {
  int num_of_alarm = 0;
  GVariantBuilder *alarm_array = nullptr;
  GVariant* param = g_variant_new("(i)", 10);
  int ret = alarm_manager_alarm_get_list_of_ids(param, 5001, getpid(), "sender",
      alarm_array, &num_of_alarm);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_appsvc_info_n) {
  gchar *b_data = nullptr;
  GVariant* param = g_variant_new("(i)", 10);
  int ret = alarm_manager_alarm_get_appsvc_info(param, 5001, &b_data);
  EXPECT_EQ(ERR_ALARM_INVALID_ID, ret);
  if (b_data)
    free(b_data);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_noti_info_n) {
  gchar *b_data = nullptr;
  GVariant* param = g_variant_new("(i)", 10);

  int ret = alarm_manager_alarm_get_noti_info(param, 5001, &b_data);
  EXPECT_EQ(ERR_ALARM_INVALID_ID, ret);
  if (b_data)
    free(b_data);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_info_n) {
  base_info_t info;
  GVariant* param = g_variant_new("(i)", 10);
  int ret = alarm_manager_alarm_get_info(param, 5001, &info);
  EXPECT_EQ(ERR_ALARM_INVALID_ID, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_next_duetime_n) {
  time_t duetime;
  GVariant* param = g_variant_new("(i)", 10);
  int ret = alarm_manager_alarm_get_next_duetime(param, 5001, &duetime);
  EXPECT_EQ(ERR_ALARM_INVALID_ID, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_all_info_p) {
  char* db_path = nullptr;
  int ret = alarm_manager_alarm_get_all_info(5001, &db_path);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
  if (db_path)
    free(db_path);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_set_rtc_time_n) {
  GVariant* param = g_variant_new("(iiiiii)", 2019, 11, 1, 12, 12, 12);
  int ret = alarm_manager_alarm_set_rtc_time(param);
  EXPECT_EQ(ERR_ALARM_SYSTEM_FAIL, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_set_time_n) {
  GVariant* param = g_variant_new("(x)", 20190012);
  int ret = alarm_manager_alarm_set_time(param, getpid(), "sender");
  EXPECT_EQ(ERR_ALARM_SYSTEM_FAIL, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_set_time_with_propagation_delay_n) {
  GVariant* param = g_variant_new("(xxxx)", 20190012, 2949294, 292929,
      29299292);
  int ret = alarm_manager_alarm_set_time_with_propagation_delay(param, getpid(),
      "sender");
  EXPECT_EQ(ERR_ALARM_SYSTEM_FAIL, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_set_timezone_n) {
  GVariant* param = g_variant_new("(s)", "/usr/share/zoneinfo/Asia/Seoul");
  int ret = alarm_manager_alarm_set_timezone(param);
  EXPECT_EQ(ERR_ALARM_SYSTEM_FAIL, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_set_global) {
  EXPECT_CALL(GetMock<PkgmgrInfoMock>(), pkgmgrinfo_pkginfo_is_global(_, _))
      .WillRepeatedly(
          Invoke([](pkgmgrinfo_pkginfo_h, bool *is_global) -> int {
            *is_global = true;
            return 0;
          }));

  gint64 reserved = 1;
  int alarm_id;
  GVariant* param = g_variant_new("(ssiiiiiiiiiiiiixss)",
      "app_service_name", "app_service_name_mod",
      2019, 11, 1, 12, 59, 59, 400, 2020, 12, 29, 3, 1, 1, reserved,
      "reserved_service_name", "reserved_service_name_mod");
  int ret = alarm_manager_alarm_create(param, 5001, getpid(), "sender", &alarm_id);
  GVariant* param2 = g_variant_new("(ib)", alarm_id, true);
  ret = alarm_manager_alarm_set_global(param2, 5001);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmServerTest, alarm_manager_alarm_get_global_n) {
  gboolean global;
  GVariant* param = g_variant_new("(i)", 1);
  int ret = alarm_manager_alarm_get_global(param, &global);
  EXPECT_EQ(ERR_ALARM_INVALID_ID, ret);
}

