/*
 * Copyright (c) 2021 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <appsvc.h>
#include <bundle_cpp.h>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <stdio.h>
#include <unistd.h>

#include "alarm.h"
#include "aul_mock.h"
#include "dbus_mock.h"
#include "glib_mock.h"
#include "notification_mock.h"
#include "pkgmgr_info_mock.h"
#include "test_fixture.h"
#include "unittest.h"

using ::testing::_;
using ::testing::DoAll;
using ::testing::Return;
using ::testing::SetArgPointee;
using ::testing::Invoke;

namespace {

void __g_dbus_node_info_unref_fake(GDBusNodeInfo *info) {
   if (info && info->interfaces) {
     free(info->interfaces);
     free(info);
   }
  info = nullptr;
  return;
}

GDBusConnection* __g_bus_get_sync_fake(GBusType type,
    GCancellable* cancellable, GError** error) {
  GDBusConnection* con = (GDBusConnection*)g_object_new(G_TYPE_OBJECT, nullptr);

  return con;
}

GDBusProxy* __g_dbus_proxy_new_sync_fake(GDBusConnection* con,
    GDBusProxyFlags flag, GDBusInterfaceInfo* info, const gchar* name,
    const gchar* path, const gchar* interface, GCancellable* cancellable,
    GError** error) {
  GDBusProxy *proxy = (GDBusProxy*)g_object_new(G_TYPE_OBJECT, nullptr);

  return proxy;
}

GDBusNodeInfo* __g_dbus_node_info_new_for_xml_fake(const gchar* xml, GError** error) {
  GDBusNodeInfo* info = (GDBusNodeInfo*)malloc(sizeof(GDBusNodeInfo));
  if (!info)
    return nullptr;

  GDBusInterfaceInfo **interface =
    (GDBusInterfaceInfo**)malloc(sizeof(GDBusInterfaceInfo*));
  if (!interface) {
    free(info);
    return nullptr;
  }

  info->ref_count = 10;
  info->path = nullptr;
  info->interfaces = interface;
  info->nodes = nullptr;
  info->annotations = nullptr;

  return info;
}

GVariant* __g_dbus_proxy_call_sync_fake_negative(GDBusProxy* proxy, const gchar* method,
    GVariant* param, GDBusCallFlags flag, gint timeout, GCancellable* cancellable,
    GError** error) {
  GQuark quark = g_quark_from_static_string("test");
  *error = g_error_new(quark, G_DBUS_ERROR_ACCESS_DENIED, "Error");

  return nullptr;
}

GVariant* __notification_ipc_make_gvariant_from_noti_fake(notification_h,
    bool translate) {
  GVariant* noti_gv = g_variant_new_string("test");

  return noti_gv;
}

int __pkgmgrinfo_pkginfo_get_api_version_fake(pkgmgrinfo_pkginfo_h handle,
    char** version) {
  *version = strdup("3.0");
  return 0;
}

int __handler(alarm_id_t alarm_id, void* user_param) {
  return 0;
}

class Mocks : virtual public ::testing::NiceMock<AulMock>,
    virtual public ::testing::NiceMock<DbusMock>,
    virtual public ::testing::NiceMock<GlibMock>,
    virtual public ::testing::NiceMock<NotificationMock>,
    virtual public ::testing::NiceMock<PkgmgrInfoMock> {};

}  // namesapce

class AlarmLibTest : public TestFixture {
 public:
  AlarmLibTest() : TestFixture(std::make_unique<::Mocks>()) {}
  void SetUp() override {
    alarm_ = alarmmgr_create_alarm();
  }
  void TearDown() override {
    alarmmgr_free_alarm(alarm_);
  }

 protected:
  alarm_entry_t* alarm_ = nullptr;
};

TEST_F(AlarmLibTest, alarmmgr_init_p) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_new_for_xml(_, _))
      .WillRepeatedly(Invoke(__g_dbus_node_info_new_for_xml_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_register_object(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_unregister_object(_, _))
      .WillRepeatedly(Return(true));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_own_name_on_connection(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_unref(_))
      .WillRepeatedly(Invoke(__g_dbus_node_info_unref_fake));

  int ret = alarmmgr_init("org.tizen.alarmmgrtestapp");
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
  alarmmgr_fini();
}

TEST_F(AlarmLibTest, alarmmgr_set_cb_p) {
  int ret = alarmmgr_set_cb([](alarm_id_t alarm_id, void *user_param) -> int {
        return 0;
      }, nullptr);

  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_get_time_p) {
  alarm_date_t date;
  alarm_date_t get_date;

  date.year = 2020;
  date.month = 1;
  date.day = 1;
  date.hour = 1;
  date.min = 1;
  date.sec = 1;

  int ret = alarmmgr_set_time(alarm_, date);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_get_time(alarm_, &get_date);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  EXPECT_EQ(date.year, get_date.year);
}

TEST_F(AlarmLibTest, alarmmgr_set_get_repeat_mode_p) {
  alarm_repeat_mode_t get_mode;
  int interval;

  int ret = alarmmgr_set_repeat_mode(alarm_, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_get_repeat_mode(alarm_, &get_mode, &interval);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  EXPECT_EQ(ALARM_REPEAT_MODE_REPEAT, get_mode);
}

TEST_F(AlarmLibTest, alarmmgr_set_get_type_p) {
  int get_type;

  int ret = alarmmgr_set_type(alarm_, ALARM_TYPE_VOLATILE);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_get_type(alarm_, &get_type);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  EXPECT_EQ(ALARM_TYPE_VOLATILE, get_type);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_appsvc_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<PkgmgrInfoMock>(),
      pkgmgrinfo_pkginfo_get_api_version(_, _))
      .WillRepeatedly(Invoke(__pkgmgrinfo_pkginfo_get_api_version_fake));

  tizen_base::Bundle b;
  int ret = appsvc_set_pkgname(b.GetHandle(),"org.tizen.setting");
  EXPECT_EQ(0, ret);

  ret = appsvc_set_operation(b.GetHandle(), APPSVC_OPERATION_DEFAULT);
  EXPECT_EQ(0, ret);

  alarm_id_t alarm_id;
  ret = alarmmgr_add_alarm_appsvc(ALARM_TYPE_DEFAULT, 100, 0,
      b.GetHandle(), &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_get_alarm_appsvc_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  int ret = -1;
  alarm_id_t alarm_id = 10;
  bundle* get_b = static_cast<bundle*>(alarmmgr_get_alarm_appsvc_info(alarm_id,
      &ret));
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  bundle_free(get_b);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_noti_n) {
  EXPECT_CALL(GetMock<GlibMock>(), g_variant_get_size(_))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<NotificationMock>(),
      notification_ipc_make_gvariant_from_noti(_, _))
      .WillRepeatedly(Invoke(__notification_ipc_make_gvariant_from_noti_fake));

  alarm_id_t alarm_id;
  notification_h noti = nullptr;
  int ret = alarmmgr_add_alarm_noti(ALARM_TYPE_DEFAULT, 100, 0, noti,
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_get_alarm_noti_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  notification_h get_noti = nullptr;
  int ret = alarmmgr_get_alarm_noti_info(1, &get_noti);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_rtc_time_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  alarm_date_t date = {2020, 4, 5, 10, 10, 0};
  int ret = alarmmgr_set_rtc_time(&date);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_with_localtime_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_new_for_xml(_, _))
      .WillRepeatedly(Invoke(__g_dbus_node_info_new_for_xml_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_register_object(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_unregister_object(_, _))
      .WillRepeatedly(Return(true));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_own_name_on_connection(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_unref(_))
      .WillRepeatedly(Invoke(__g_dbus_node_info_unref_fake));

  alarm_id_t alarm_id;
  alarm_date_t test_time;

  test_time.year = 2022;
  test_time.month = 1;
  test_time.day = 1;
  test_time.hour = 1;
  test_time.min = 1;
  test_time.sec = 5;

  alarmmgr_fini();
  int ret = alarmmgr_init("org.tizen.alarmmgrtestapp");
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_time(alarm_,test_time);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_type(alarm_, ALARM_TYPE_DEFAULT);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_repeat_mode(alarm_, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_add_alarm_with_localtime(alarm_, "org.tizen.alarmmgrtestapp",
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  alarmmgr_fini();
}

TEST_F(AlarmLibTest, alarmmgr_get_info_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  alarm_entry_t get_alarm;
  int ret = alarmmgr_get_info(1, &get_alarm);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_appsvc_with_localtime_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  tizen_base::Bundle b;
  int ret = appsvc_set_pkgname(b.GetHandle(),"org.tizen.calendar");
  EXPECT_EQ(0, ret);

  ret = appsvc_set_operation(b.GetHandle(), APPSVC_OPERATION_DEFAULT);
  EXPECT_EQ(0, ret);

  alarm_id_t alarm_id;
  alarm_date_t test_time;

  test_time.year = 2022;
  test_time.month = 1;
  test_time.day = 1;
  test_time.hour = 1;
  test_time.min = 1;
  test_time.sec = 5;

  ret = alarmmgr_set_time(alarm_,test_time);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_set_type(alarm_, ALARM_TYPE_DEFAULT);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_add_alarm_appsvc_with_localtime(alarm_, b.GetHandle(),
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_noti_with_localtime_n) {
  EXPECT_CALL(GetMock<GlibMock>(), g_variant_get_size(_))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<NotificationMock>(),
      notification_ipc_make_gvariant_from_noti(_, _))
      .WillRepeatedly(Invoke(__notification_ipc_make_gvariant_from_noti_fake));

  int ret = alarmmgr_set_repeat_mode(alarm_, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  alarm_id_t alarm_id;
  notification_h noti = nullptr;
  ret = alarmmgr_add_alarm_noti_with_localtime(alarm_, noti, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  int ret = alarmmgr_set_repeat_mode(alarm_, ALARM_REPEAT_MODE_REPEAT, 1000);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  alarm_id_t alarm_id;
  ret = alarmmgr_add_alarm(ALARM_TYPE_VOLATILE, 100, 1000, "gtest-alarmmgr",
      &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_alarm_precision_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  alarm_id_t alarm_id;
  int ret = alarmmgr_add_alarm_precision(ALARM_TYPE_VOLATILE, 100, 1000,
      "gtest-alarmmgr", &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_get_alarm_withcb_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_new_for_xml(_, _))
      .WillRepeatedly(Invoke(__g_dbus_node_info_new_for_xml_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_register_object(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_unregister_object(_, _))
      .WillRepeatedly(Return(true));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_own_name_on_connection(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_unref(_))
      .WillRepeatedly(Invoke(__g_dbus_node_info_unref_fake));

  alarmmgr_fini();
  alarm_id_t alarm_id;
  int ret = alarmmgr_add_alarm_withcb(ALARM_TYPE_VOLATILE, 100, 1000,
      __handler, nullptr, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  alarmmgr_fini();
}

TEST_F(AlarmLibTest, alarmmgr_add_get_alarm_withcb_precision_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_new_for_xml(_, _))
      .WillRepeatedly(Invoke(__g_dbus_node_info_new_for_xml_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_register_object(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_unregister_object(_, _))
      .WillRepeatedly(Return(true));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_own_name_on_connection(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_unref(_))
      .WillRepeatedly(Invoke(__g_dbus_node_info_unref_fake));

  alarm_id_t alarm_id;
  int ret = alarmmgr_add_alarm_withcb_precision(ALARM_TYPE_VOLATILE, 100, 1000,
      __handler, nullptr, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  alarmmgr_fini();
}

TEST_F(AlarmLibTest, alarmmgr_remove_alarm_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  int ret = alarmmgr_remove_alarm(1);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_remove_all_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  int ret = alarmmgr_remove_all();
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_enum_alarm_ids_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  int ret = alarmmgr_enum_alarm_ids([](alarm_id_t alarm_id,
      void *user_param) -> int {
        return 0;
      }, nullptr);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_get_next_duetime_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  time_t next_time;
  int ret = alarmmgr_get_next_duetime(1, &next_time);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_add_periodic_alarm_withcb_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_new_for_xml(_, _))
      .WillRepeatedly(Invoke(__g_dbus_node_info_new_for_xml_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_register_object(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_unregister_object(_, _))
      .WillRepeatedly(Return(true));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_own_name_on_connection(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_unref(_))
      .WillRepeatedly(Invoke(__g_dbus_node_info_unref_fake));

  alarm_id_t alarm_id;
  int ret = alarmmgr_add_periodic_alarm_withcb(1000, CUT_OFF,
      [](alarm_id_t alarm_id, void*) -> int {
        return 0;
      }, nullptr, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  alarmmgr_fini();
}

TEST_F(AlarmLibTest, alarmmgr_add_reference_periodic_alarm_withcb_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_new_for_xml(_, _))
      .WillRepeatedly(Invoke(__g_dbus_node_info_new_for_xml_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_register_object(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_connection_unregister_object(_, _))
      .WillRepeatedly(Return(true));
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_own_name_on_connection(_, _, _,
      _, _, _, _))
      .WillRepeatedly(Return(1));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_node_info_unref(_))
      .WillRepeatedly(Invoke(__g_dbus_node_info_unref_fake));

  alarm_id_t alarm_id;
  int ret = alarmmgr_add_reference_periodic_alarm_withcb(1000,
      [](alarm_id_t alarm_id, void*) -> int {
        return 0;
      }, nullptr, &alarm_id);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
  alarmmgr_fini();
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  time_t current_time;
  time(&current_time);
  int ret = alarmmgr_set_systime(current_time);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_with_propagation_delay_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  struct timespec newtime;
  struct timespec reqtime;
  clock_gettime(CLOCK_REALTIME, &newtime);
  clock_gettime(CLOCK_REALTIME, &reqtime);
  int ret = alarmmgr_set_systime_with_propagation_delay(newtime, reqtime);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_async_p) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  time_t current_time;
  time(&current_time);
  int ret = alarmmgr_set_systime_async(current_time,
      [](int result, void*) -> int {
        return 0;
      }, nullptr);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_systime_with_propagation_delay_async_p) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  struct timespec newtime;
  struct timespec reqtime;
  clock_gettime(CLOCK_REALTIME, &newtime);
  clock_gettime(CLOCK_REALTIME, &reqtime);
  newtime.tv_sec += 100;
  reqtime.tv_sec += 100;
  int ret = alarmmgr_set_systime_with_propagation_delay_async(newtime, reqtime,
      [](int result, void*) -> int {
        return 0;
      },nullptr);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_timezone_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  char zone[] = "/usr/share/zoneinfo/Asia/Seoul";
  int ret = alarmmgr_set_timezone(zone);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_set_global_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  int ret = alarmmgr_set_global(1, true);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_get_global_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  bool global;
  int ret = alarmmgr_get_global(1, &global);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}

TEST_F(AlarmLibTest, alarmmgr_update_alarm_n) {
  EXPECT_CALL(GetMock<DbusMock>(), g_bus_get_sync(_, _, _))
      .WillRepeatedly(Invoke(__g_bus_get_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_new_sync(
      _, _, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_new_sync_fake));
  EXPECT_CALL(GetMock<DbusMock>(), g_dbus_proxy_call_sync(_, _, _, _, _, _, _))
      .WillRepeatedly(Invoke(__g_dbus_proxy_call_sync_fake_negative));

  alarm_date_t test_time;
  test_time.year = 2022;
  test_time.month = 1;
  test_time.day = 1;
  test_time.hour = 1;
  test_time.min = 1;
  test_time.sec = 5;

  int ret = alarmmgr_set_time(alarm_, test_time);
  EXPECT_EQ(ALARMMGR_RESULT_SUCCESS, ret);

  ret = alarmmgr_update_alarm(1, alarm_, ALARM_UPDATE_FLAG_TIME);
  EXPECT_EQ(ERR_ALARM_NO_PERMISSION, ret);
}
